import time
from unittest import TestCase

from pydis import Pydis
from pydis.exceptions import NotFound


class TestCore(TestCase):
    def test_set_get(self):
        p = Pydis()
        fake_key, fake_value = 'fake_key', 'fake_value'
        p.set(fake_key, fake_value)
        ret = p.get(fake_key)
        self.assertEqual(ret, fake_value)

    def test_ttl(self):
        p = Pydis()
        fake_key, fake_value = 'timeout_key', 'timeout_value'
        p.set(fake_key, fake_value, timeout=2)
        self.assertEqual(p.get(fake_key), fake_value)
        self.assertIs(p.ttl(fake_key) > 0, True)
        time.sleep(2)
        self.assertIsNone(p.get(fake_key))

    def test_nx(self):
        p = Pydis()
        fake_key, fake_value = 'nx_key', 'nx_value'
        p.set_nx(fake_key, fake_value)
        self.assertEqual(p.get(fake_key), fake_value)
        p.set_nx(fake_key, "new_value")
        self.assertEqual(p.get(fake_key), fake_value)

    def test_delete(self):
        p = Pydis()
        fake_key, fake_value = 'delete_key', 'delete_value'
        p.set(fake_key, fake_value)
        p.delete(fake_key)
        self.assertIsNone(p.get(fake_key))

    def test_incr(self):
        p = Pydis()
        fake_key = 'incr_key'
        p.set(fake_key, 0)
        p.incr(fake_key)
        self.assertEqual(p.get(fake_key), 1)

        p.incr(fake_key, 2)
        self.assertEqual(p.get(fake_key), 3)

        with self.assertRaises(NotFound):
            p.incr('dose not exists key')

    def test_decr(self):
        p = Pydis()
        fake_key = 'decr_key'
        p.set(fake_key, 3)
        p.decr(fake_key)
        self.assertEqual(p.get(fake_key), 2)

        p.decr(fake_key, 2)
        self.assertEqual(p.get(fake_key), 0)

        with self.assertRaises(NotFound):
            p.incr('dose not exists key')

    def test_keys(self):
        p = Pydis()
        p.force_clean()
        keys = ['e', 'a', 'b', 'c', 'd']
        for index, key in enumerate(keys, 1):
            p.set(key, 1, index)

        time.sleep(2)
        self.assertListEqual(['b', 'c', 'd'], p.keys())

    def test_clean(self):
        p = Pydis()
        p.set('a', 1, 1)
        p.set('b', 2)

        time.sleep(1)
        p.clean()
        with self.assertRaises(NotFound):
            _ = p._data['a']
        self.assertEqual(p.get('b'), 2)

    def test_force_clean(self):
        p = Pydis()
        keys = ['e', 'a', 'b', 'c', 'd']
        for key in keys:
            p.set(key, 1)
        p.force_clean()
        self.assertListEqual([], p.keys())

    def test_len(self):
        p = Pydis()
        p.force_clean()
        self.assertEqual(len(p), 0)



class PydisError(Exception):
    detail = 'An internal error occurred'

    def __init__(self, detail=None):
        if detail is not None:
            self.detail = detail

    def __str__(self):
        return str(self.detail)


class NotFound(PydisError):
    detail = 'Key : `%s` does not exists.'

    def __init__(self, key):
        self.detail = self.detail % key


class ExpiredError(NotFound):
    detail = 'Key : `%s` already expired and deleted.'



from typing import Any, Optional, Tuple
from datetime import datetime, timedelta


class Value:
    def __init__(self, value: Any, timeout: Optional[float]):
        self.is_forever, self.expired_to = self.__load_expired_to(timeout)
        self.can_incr = isinstance(value, int)
        self.value = value

    @staticmethod
    def __load_expired_to(timeout: Optional[float]) -> Tuple[bool, Optional[datetime]]:
        if timeout is None:
            is_forever = True
            expired_to = None
        else:
            if timeout <= 0:
                raise ValueError("Make sure timeout is an float greater than 0 ")
            is_forever = False
            expired_to = datetime.now() + timedelta(seconds=timeout)
        return is_forever, expired_to

    def is_expired(self) -> bool:
        if self.is_forever:
            return False
        return self.expired_to < datetime.now()

    def ttl(self) -> int:
        if self.is_forever:
            return -1
        return (self.expired_to - datetime.now()).seconds

    def incr(self, amplitude: int, func: str) -> None:
        if not self.can_incr:
            raise ValueError(f"type: {type(self.value)} no support {func}")
        self.value += amplitude



import threading

from .exceptions import NotFound, ExpiredError


class SingletonType(type):
    _instance_lock = threading.Lock()
    _instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            with cls._instance_lock:
                if cls._instance is None:
                    cls._instance = super().__call__(*args, **kwargs)
        return cls._instance


class Data(dict):
    def __getitem__(self, key):
        try:
            value = super().__getitem__(key)
            if value.is_expired():
                self.pop(key)
                raise ExpiredError(key)
        except KeyError:
            raise NotFound(key)
        return value



from typing import Any, Optional, List

from .utils import SingletonType, Data
from .value import Value
from .exceptions import ExpiredError, NotFound


class Pydis(metaclass=SingletonType):
    _data: Data[str, Value] = Data()

    def __init__(self, default_timeout: Optional[int] = None):
        """
        :param default_timeout: 全局的timeout，如果在设置key没有指定timeout的话，就会应用该timeout
        """
        self.default_timeout = default_timeout

    def force_clean(self) -> None:
        """
        删除所有的keys，没有过期的key也会被删除，慎用
        :return:
        """
        self._data.clear()

    def clean(self) -> None:
        """
        删除已经过期的key
        :return:
        """
        expired_keys = [key for key, value in self._data.items() if value.is_expired()]
        _ = [self.delete(expired_key) for expired_key in expired_keys]

    def get(self, key: str) -> Any:
        try:
            value = self._data[key]
        except (NotFound, ExpiredError):
            return None
        return value.value

    def set_nx(self, key: str, value: Any, timeout: Optional[int] = None) -> bool:
        """
        没有key的时候设置key，返回True
        有key的时候不做任何操作，返回False
        """
        if self.get(key) is not None:
            return False
        self._set(key, value, timeout)
        return True

    def set(self, key: str, value: Any, timeout: Optional[int] = None) -> None:
        self._set(key, value, timeout)

    def _set(self, key: str, value: Any, timeout: Optional[int] = None) -> None:
        if timeout is None:
            timeout = self.default_timeout
        value = Value(value, timeout=timeout)
        self._data[key] = value

    def delete(self, key: str) -> None:
        self._data.pop(key, None)

    def ttl(self, key: str) -> int:
        """
        返回持续时间，如果key不存在返回-2，永久key返回-1，其他情况，返回具体的到期时间
        :param key:
        :return:
        """
        try:
            value = self._data[key]
        except (NotFound, ExpiredError):
            return -2
        return value.ttl()

    def _incr(self, key: str, amplitude: int, func: str) -> int:
        value = self._data[key]
        value.incr(amplitude, func)
        return value.value

    def incr(self, key: str, amplitude: int = 1) -> int:
        return self._incr(key, amplitude, 'incr')

    def decr(self, key: str, amplitude: int = 1) -> int:
        return self._incr(key, -amplitude, 'decr')

    def keys(self) -> List:
        """
        以列表形式返回所有的key
        :return:
        """
        keys = []
        expired_keys = []
        for key, value in self._data.items():
            if value.is_expired():
                expired_keys.append(key)
            else:
                keys.append(key)

        _ = [self.delete(expired_key) for expired_key in expired_keys]
        return keys

    def is_empty(self) -> bool:
        return bool(self._data)

    def __len__(self):
        return len(self._data)
